@@ -1,11 +1,13 @@
 """ This module has routines for analysing the absorption profiles
 from ions and molecules.
 """
-import voigt
+from __future__ import division
+from voigt import voigt
 from convolve import convolve_psf
 from utilities import between, adict, get_data_path, indexnear
 from constants import Ar, me, mp, kboltz, c, e, sqrt_ln2, c_kms
-from spec import find_wa_edges, make_constant_dv_wa_scale
+from spec import find_wa_edges
+from sed import  make_constant_dv_wa_scale
 from abundances import Asolar
 from pyvpfit import readf26
 
@@ -13,16 +15,18 @@ import numpy as np
 
 from cStringIO import StringIO
 import math
-from math import pi, sqrt
+from math import pi, sqrt, exp
 
 
 DATAPATH = get_data_path()
 
-def calctau(v, wav0, osc, gam, logN, T=None, btemp=20, bturb=0,
-            debug=False, verbose=True):
+# this constant gets used in several functions (units of cm^2/s)
+e2_me_c = e**2 / (me*c)
+
+def calctau(vel, wa0, osc, gam, logN, b, debug=False, verbose=True):
     """ Returns the optical depth (Voigt profile) for a transition.
 
-    Given an transition with rest wavelength wav0, osc strength,
+    Given an transition with rest wavelength wa0, osc strength,
     natural linewidth gam; b parameter (doppler and turbulent); and
     log10 (column density), returns the optical depth in velocity
     space. v is an array of velocity values in km/s. The absorption
@@ -30,104 +34,140 @@ def calctau(v, wav0, osc, gam, logN, T=None, btemp=20, bturb=0,
 
     Parameters
     ----------
-    v : array of floats, shape (N,)
+    vel : array of floats, shape (N,)
       Velocities in km/s.
-    wav0 : float
-      Rest wavelength op transition in Angstroms.
+    wa0 : float
+      Rest wavelength of transition in Angstroms.
     osc : float
       Oscillator strength of transition (dimensionless).
     gam : float
       Gamma parameter for the transition (dimensionless).
     logN : float:
       log10 of the column density in absorbers per cm^2.
-    btemp : float (20)
-      *b* parameter from doppler temperature broadening (km/s)
-    bturb : float (0)
-      *b* parameter from doppler turbulent broadening (km/s)
-    T : float (None)
-      Temperature of cloud in Kelvin (overrides btemp).
+    b : float
+      *b* parameter (km/s).
 
     Returns
     -------
     tau : array of floats, shape (N,)
-      The optical depth as a function of `v`.
+      The optical depth as a function of `vel`.
 
     Notes
     -----
-    The step size for `v` must be small enough to properly
+    The step size for `vel` must be small enough to properly
     sample the profile.
 
     To map the velocity array to some wavelength for a transitions
-    with rest wavelength wav0 at redshift z:
+    with rest wavelength wa0 at redshift z:
 
     >>> z = 3.0
-    >>> wa = wav0 * (1 + z) * (1 + v/c_kms)
+    >>> wa = wa0 * (1 + z) * (1 + v/c_kms)
     """
     # note units are cgs
-    wav0 = wav0 * 1e-8                    # cm
+    wa0 = wa0 * 1e-8                    # cm
     N = 10**logN                          # absorbers/cm^2
-    if T is not None:
-        btemp = sqrt(2*kboltz*T / mp) * 1e-5     # km/s
-    b = math.hypot(btemp, bturb) * 1e5    # cm/s
-    nu0 = c / wav0                        # rest frequency, s^-1
+    b = b * 1e5                           # cm/s
+    nu0 = c / wa0                        # rest frequency, s^-1
     # Now use doppler relation between v and nu assuming gam << nu0
     gam_v = gam / nu0 * c             # cm/s
     if debug:
         print ('Widths in km/s (Lorentzian Gamma, Gaussian b):',
                gam_v/1.e5, b/1.e5)
 
-    fwhml = gam_v / (2.*pi)                # cm/s
-    fwhmg = 2. * sqrt_ln2 * b                # cm/s
+    if verbose:
+        fwhml = gam_v / (2.*pi)                # cm/s
+        fwhmg = 2. * sqrt_ln2 * b                # cm/s
 
-    ##### sampling check, first get velocity width ######
-    ic = np.searchsorted(v, 0)
-    try:
-        # it's ok if the transition is outside the fitting region.
-        if ic == len(v):
-            ic -= 1
-        if ic == 0:
-            ic += 1
-        vstep = v[ic] - v[ic-1]
-    except IndexError:
-        raise IndexError(4*'%s ' % (len(v), ic, ic-1, v))
+        ##### sampling check, first get velocity width ######
+        ic = np.searchsorted(vel, 0)
+        try:
+            # it's ok if the transition is outside the fitting region.
+            if ic == len(vel):
+                ic -= 1
+            if ic == 0:
+                ic += 1
+            vstep = vel[ic] - vel[ic-1]
+        except IndexError:
+            raise IndexError(4*'%s ' % (len(vel), ic, ic-1, vel))
 
-    fwhm = max(gam_v/1.e5, fwhmg/1.e5)
+        fwhm = max(gam_v/1.e5, fwhmg/1.e5)
+        
+        if vstep > fwhm:
+            print 'Warning: tau profile undersampled!'
+            print '  Pixel width: %f km/s, transition fwhm: %f km/s' % (vstep,fwhm)
+            # best not to correct for this here, because even if we do,
+            # we'll get nonsense if we convolve the resulting flux with an
+            # instrumental profile.  Need to use smaller dv size
+            # throughout tau, exp(-tau) and convolution of exp(-tau)
+            # calculations, only re-binning back to original dv size after
+            # all these steps.
 
-    if verbose and vstep > fwhm:
-        print 'Warning: tau profile undersampled!'
-        print '  Pixel width: %f km/s, transition fwhm: %f km/s' % (vstep,fwhm)
-        # best not to correct for this here, because even if we do,
-        # we'll get nonsense if we convolve the resulting flux with an
-        # instrumental profile.  Need to use smaller dv size
-        # throughout tau, exp(-tau) and convolution of exp(-tau)
-        # calculations, only re-binning back to original dv size after
-        # all these steps.
-
-    u = 1.e5 / b * v                           # dimensionless
+    u = 1.e5 / b * vel                         # dimensionless
     a = gam_v / (4*pi*b)                       # dimensionless
-    vp = voigt.voigt(a, u)                     # dimensionless
-    const = pi * e**2 / (me*c)                 # m^2/s
-    tau = const*N*osc*wav0 / (sqrt(pi)*b) * vp # dimensionless
+    vp = voigt(a, u)                           # dimensionless
+    tau = pi * e2_me_c * N * osc * wa0 / (sqrt(pi) * b) * vp # dimensionless
 
     return tau
 
-def calc_tau_peak(logN, b, trans):
-    """
-    Estimate of  the peak optical depth of a transition assuming we are on the
-    linear part of the curve of growth.
+def calc_tau_peak(logN, b, wa0, osc):
+    """ Find the optical depth of a transition at line centre assuming
+    we are on the linear part of the curve of growth.
 
-    logN is log10 of column density in cm^-2
-    b is b parameter in km/s
-    trans is  a transisition, e.e.g 'HI 1215'
+    Parameters
+    ----------
+    logN : array_like
+      log10 of column density in cm^-2
+    b : float
+      b parameter in km/s.
+    wa0 : float
+      Rest wavelength of the transition in Angstroms.
+    osc : float
+      Transition oscillator strength.
+
+    Returns
+    -------
+    tau0 : ndarray or scalar if logN is scalar
+      optical depth at the centre of the line
+
+    Notes
+    -----
+    See Draine "Physics of the Interstellar and Intergalactic medium"
+    for a description of how to calculate this quantity.
     """
     b_cm_s = b * 1e5
-    if isinstance(trans, basestring):
-        tname, trans = findtrans(trans, readatom())
-    wav0 = trans[0] * 1e-8 # cm
-    osc = trans[1]
-    const = pi * e**2 / (me*c)
-    #gauss_norm = 1 / (sigma * sqrt(2*pi))
-    return 10**logN * osc * wav0 * const / (b_cm_s * sqrt(pi))
+
+    wa0 = wa0 * 1e-8 # cm
+    return sqrt(pi) * e2_me_c * 10**logN * osc * wa0 / b_cm_s
+
+def logN_from_tau_peak(tau, b, wa0, osc):
+    """ Calculate the column density for a transition given its
+    width and its optical depth at line centre.
+    
+    Parameters
+    ----------
+    tau : array_like
+      Optical depth at line centre.
+    b : float
+      b parameter in km/s.
+    wa0 : float
+      Rest wavelength of the transition in Angstroms.
+    osc : float
+      Transition oscillator strength.
+
+    Returns
+    -------
+    logN : ndarray, or scalar if tau is scalar
+      log10 of column density in cm^-2
+
+    See Also
+    --------
+    calc_tau_peak
+    """
+    b_cm_s = b * 1e5
+
+    wa0 = wa0 * 1e-8 # cm
+
+    return np.log10(tau * b_cm_s / (sqrt(pi) * e2_me_c * osc * wa0))
 
 
 def calc_iontau(wa, ion, zp1, logN, b, debug=False, ticks=False, maxdv=1000.,
@@ -178,15 +218,18 @@ def calc_iontau(wa, ion, zp1, logN, b, debug=False, ticks=False, maxdv=1000.,
     tickmarks = []
     sumtau = np.zeros_like(wa)
     i0 = i1 = None 
-    for i,(wav0,osc,gam) in enumerate(trans):
-        refwav = wav0 * zp1
+    for i,(wa0,osc,gam) in enumerate(trans):
+        tau0 = calc_tau_peak(logN, b, wa0, osc)
+        if 1 - exp(-tau0) < 1e-3:
+            continue
+        refwav = wa0 * zp1
         dv = (wa - refwav) / refwav * c_kms
         if maxdv is not None:
             i0,i1 = dv.searchsorted([-maxdv, maxdv])
-        tau = calctau(dv[i0:i1], wav0, osc, gam, logN, btemp=b,
+        tau = calctau(dv[i0:i1], wa0, osc, gam, logN, b,
                       debug=debug, verbose=verbose)
-        if ticks and tau.max() > label_tau_threshold:
-            tickmarks.append((refwav, z, wav0, i))
+        if ticks and tau0 > label_tau_threshold:
+            tickmarks.append((refwav, z, wa0, i))
         sumtau[i0:i1] += tau
 
     if ticks:
@@ -200,7 +243,7 @@ def find_tau(wa, lines, atom, per_trans=False):
     log10(column density) and b parameter, return the tau at each
     wavelength from all these transitions.
 
-    lines can also be the name of a fort.26 format file
+    lines can also be the name of a VPFIT fort.26 format file.
 
     Note this assumes the wavelength array has small enough pixel
     separations so that the profiles are properly sampled.
@@ -221,7 +264,8 @@ def find_tau(wa, lines, atom, per_trans=False):
     for ion,z,b,logN in lines:
         #print 'z, logN, b', z, logN, b
         maxdv = 20000 if logN > 18 else 1000
-        t,tick = calc_iontau(wa, atom[ion], z+1, logN, b, ticks=True,maxdv=maxdv)
+        t,tick = calc_iontau(wa, atom[ion], z+1, logN, b, ticks=True,
+                             maxdv=maxdv)
         tau += t
         if per_trans:
             taus.append(t)
@@ -328,7 +372,9 @@ def calc_Wr(i0, i1, wa, tr, ew=None, ewer=None, fl=None, er=None, co=None,
         Wrelo = Welo / zp1
     
     # Assume we are on the linear part of curve of growth (will be an
-    # underestimate if saturated)
+    # underestimate if saturated). See Draine, "Physics of the
+    # Interstellar and Intergalactic medium", ISBN 978-0-691-12214-4,
+    # chapter 9.
     Nmult = 1.13e20 / (tr['osc'] * tr['wa']**2)
 
     # 5 sigma detection limit
@@ -378,7 +424,7 @@ def b_to_T(atom, bvals):
     return  T
 
 def T_to_b(atom, T):
-    """ Convert temperatues in K to b parameters (km/s) for an atom
+    """ Convert temperatures in K to b parameters (km/s) for an atom
     with mass amu.
 
     Parameters
@@ -420,8 +466,8 @@ def read_HITRAN(thelot=False):
 
     The returned wavelengths are in Angstroms.
 
-    The strongest absorption features in the optical range is
-    typically O2.
+    The strongest absorption features in the optical range are
+    typically due to O2.
     """
     filename = DATAPATH + '/linelists/HITRAN2004_wa_lt_25000.fits.gz'
     lines = readtabfits(filename)
@@ -476,10 +522,11 @@ def readatom(filename=None, debug=False,
 
     atom = dict()
     atomflat = []
-
+    specials = set(['??', '__', '>>', '<<', '<>'])
     for line in fh:
         if debug:  print line
-        if not line[0].isupper():  continue
+        if not line[0].isupper() and line[:2] not in specials:
+            continue
         ion = line[:6].replace(' ','')
         if not molecules:
             if ion[:2] in set(['HD','CO','H2']):
@@ -539,7 +586,7 @@ def split_trans_name(name):
         i += 1
     return name[:i], name[i:]
 
-def tau_LL(logN, wa, wstart=912.):
+def tau_LL(logN, wa, wstart=912):
     """ Find the optical depth at the neutral hydrogen Lyman limit.
 
     Parameters
@@ -549,7 +596,7 @@ def tau_LL(logN, wa, wstart=912.):
     wa : array_like
       Wavelength in Angstroms.
     wstart : float (912.)
-      Tau values at wavelengths above this are zero.
+      Tau values at wavelengths above this are set to zero.
 
     Returns
     -------
@@ -600,8 +647,9 @@ def tau_LL(logN, wa, wstart=912.):
     tau[:i] = 10**logN * sigma0 * (wa[:i] / 912.)**3 
     return tau
 
-def calc_DLA_tau(wmin, wmax, logN=20.3, logZ=0, dv=5., atom=None):
-    """ Create the optical due to absorption from a DLA.
+def calc_DLA_tau(wa, z=0, logN=20.3, logZ=0, bHI=50, atom=None,
+                 verbose=1, highions=1, molecules=False):
+    """ Create the optical depth due to absorption from a DLA.
 
     The DLA is at z=0. The column density and metallicity can be
     varied. Solar Abundance ratios are assumed, with most of the atoms
@@ -609,49 +657,73 @@ def calc_DLA_tau(wmin, wmax, logN=20.3, logZ=0, dv=5., atom=None):
 
     Parameters
     ----------
-    wmin, wmax : float
-       Start an end of the wavelength scale (rest wavelengths).
+    wa : array_like
+       Wavelength scale.
     logZ : float (0)
        log10 of the metal abundance relative to solar. For example 0
        (the default) gives solar abundances, -1 gives 1/10th of solar.
-       
+    verbose : bool (False)
+       Print helpful information
+
     Returns
     -------
-    wa, tau : ndarrays, shape (N,)
-      The wavelength array and tau at each wavelength.
-
+    tau, ticks : ndarrays, structured array
+      tau at each wavelength and tick positions and names.
     """
-    wa = make_constant_dv_wa_scale(wmin, wmax, dv)
     off = -12 + logN + logZ
-    f26 = (
-        'HI    0  0  50 0 %.2f 0' % logN,    
-        'OI    0  0  50 0 %.2f 0' % (Asolar['O']  + off), 
-        'SiII  0  0  50 0 %.2f 0' % (Asolar['Si'] + off - 0.05), 
-        'SiIII 0  0  50 0 %.2f 0' % (Asolar['Si'] + off - 1),    
-        'SiIV  0  0  50 0 %.2f 0' % (Asolar['Si'] + off - 1),    
-        'FeII  0  0  50 0 %.2f 0' % (Asolar['Fe'] + off),    
-        'CII   0  0  50 0 %.2f 0' % (Asolar['C']  + off - 0.05),    
-        'CIII  0  0  50 0 %.2f 0' % (Asolar['C']  + off - 1),    
-        'CIV   0  0  50 0 %.2f 0' % (Asolar['C']  + off - 1),
-        'CaII  0  0  50 0 %.2f 0' % (Asolar['Ca'] + off),
-        'AlII  0  0  50 0 %.2f 0' % (Asolar['Al'] + off - 0.05),
-        'AlIII 0  0  50 0 %.2f 0' % (Asolar['Al'] + off - 1),
-        'TiII  0  0  50 0 %.2f 0' % (Asolar['Ti'] + off),
-        'NII   0  0  50 0 %.2f 0' % (Asolar['N']  + off),
-        'ZnII  0  0  50 0 %.2f 0' % (Asolar['Zn']  + off),
-        'CrII  0  0  50 0 %.2f 0' % (Asolar['Cr']  + off),
-        )
+    temp = b_to_T('H', bHI)
+    print 'b %.2f km/s gives a temperature of %.1f K' % (bHI, temp)
+    
+    elements = 'O Si Fe C Ca Al Ti N Zn Cr'.split()
+    b = dict((el, T_to_b(el, temp)) for el in elements)
+    print 'using low ion b values:'
+    print ', '.join('%s %.1f' % (el, b[el]) for el in elements)
+    
+    f26 = [
+        'HI    %.6f  0  %.2f 0 %.2f 0' % (z, bHI, logN),    
+        'OI    %.6f  0  %.2f 0 %.2f 0' % (z, b['O'], (Asolar['O']  + off)), 
+        'SiII  %.6f  0  %.2f 0 %.2f 0' % (z, b['Si'], (Asolar['Si'] + off - 0.05)), 
+        'SiIII %.6f  0  %.2f 0 %.2f 0' % (z, b['Si'], (Asolar['Si'] + off - 1)),    
+        'FeII  %.6f  0  %.2f 0 %.2f 0' % (z, b['Fe'], (Asolar['Fe'] + off)),    
+        'CII   %.6f  0  %.2f 0 %.2f 0' % (z, b['C'], (Asolar['C']  + off - 0.05)),    
+        'CIII  %.6f  0  %.2f 0 %.2f 0' % (z, b['C'], (Asolar['C']  + off - 1)),    
+        'CaII  %.6f  0  %.2f 0 %.2f 0' % (z, b['Ca'], (Asolar['Ca'] + off)),
+        'AlII  %.6f  0  %.2f 0 %.2f 0' % (z, b['Al'], (Asolar['Al'] + off - 0.05)),
+        'AlIII %.6f  0  %.2f 0 %.2f 0' % (z, b['Al'], (Asolar['Al'] + off - 1)),
+        'TiII  %.6f  0  %.2f 0 %.2f 0' % (z, b['Ti'], (Asolar['Ti'] + off)),
+        'NII   %.6f  0  %.2f 0 %.2f 0' % (z, b['N'], (Asolar['N']  + off)),
+        'ZnII  %.6f  0  %.2f 0 %.2f 0' % (z, b['Zn'], (Asolar['Zn']  + off)),
+        'CrII  %.6f  0  %.2f 0 %.2f 0' % (z, b['Cr'], (Asolar['Cr']  + off)),
+        ]
+    if highions:
+        print 'Including O VI, Si IV, C IV and N V'
+        logNCIV = 15.
+        logNSiIV = logNCIV - (Asolar['C'] - Asolar['Si'] )
+        f26 = f26 + [
+            'OVI   %.6f  0  30.0 0 15.0 0' % z,
+            'SiIV  %.6f  0  %.2f  0 %.2f 0' % (z, b['Si'], logNSiIV),
+            'CIV   %.6f  0  %.2f 0 %.2f 0' % (z, b['C'], logNCIV),
+            'NV    %.6f  0  30.0 0 15.0 0' % z,
+            ]
+    if molecules:
+        print 'Including H_2 and CO'
+        f26 = f26 + [
+            'H2J0  %.6f  0  5.0  0 18.0 0' % z,
+            'H2J1  %.6f  0  5.0  0 19.0 0' % z,
+            'COJ0  %.6f  0  5.0  0 14.0 0' % z,
+            ]
+
     f26 = StringIO('\n'.join(f26))
     if atom is None:
-        atom = readatom()
+        atom = readatom(molecules=molecules)
 
-    #import pdb; pdb.set_trace()
     tau,ticks = find_tau(wa, f26, atom)
-    tau += tau_LL(logN, wa, wstart=912.)
-    return wa, tau, ticks
+    tau += tau_LL(logN, wa/(1+z), wstart=912.5)
+    return tau, ticks
 
-def calc_DLA_trans(wa, redshift, vfwhm, logN=20.3, logZ=0, dv=5.):
-    """ Find the transmission after absrption by a DLA
+def calc_DLA_trans(wa, redshift, vfwhm, logN=20.3, logZ=0, bHI=50,
+                   highions=True, molecules=False):
+    """ Find the transmission after absorption by a DLA
 
     Parameters
     ----------
@@ -664,6 +736,9 @@ def calc_DLA_trans(wa, redshift, vfwhm, logN=20.3, logZ=0, dv=5.):
     logZ : float (0)
        log10 of the metal abundance relative to solar. For example 0
        (the default) gives solar abundances, -1 gives 1/10th of solar.
+    bHI : float (40)
+       b parameter for the HI. Other species are assumed to be
+       thermally broadened.
 
     Returns
     -------
@@ -672,10 +747,33 @@ def calc_DLA_trans(wa, redshift, vfwhm, logN=20.3, logZ=0, dv=5.):
     ticks : ndarray, shape (M,)
       Information for making ticks to show absorbing components.
     """
-    wa1 = wa / (1 + redshift)
-    lwa1, tau, ticks = calc_DLA_tau(wa1[0], wa1[-1], logN=logN, logZ=logZ,
-                                    dv=dv)
-    lwa = lwa1 * (1 + redshift)
-    trans = convolve_psf(np.exp(-tau),  vfwhm / dv)
-    trans1 = np.interp(wa, lwa, trans)
+    wa1 = make_constant_dv_wa_scale(wa[0], wa[-1], vfwhm/3.)  
+    tau, ticks = calc_DLA_tau(
+        wa1, z=redshift, logN=logN, logZ=logZ, bHI=bHI,
+        highions=highions, molecules=molecules)
+    trans = convolve_psf(np.exp(-tau),  3.)
+    trans1 = np.interp(wa, wa1, trans)
     return trans1, ticks
+
+
+def guess_logN_b(ion, wa0, osc, tau0):
+    """ Estimate logN and b for a transition given the peak optical
+    depth and atom.
+
+    Examples
+    --------
+    >>> logN, b = guess_b_logN('HI', 1215.6701, nfl, ner)
+    """
+
+    T = 10000
+    if ion == 'HI':
+        T = 30000
+    elif ion in ('CIV', 'SiIV'):
+        T = 20000
+        
+    b = T_to_b(split_trans_name(ion)[0], T)
+
+    if ion == 'OVI':
+        b = 30.
+
+    return logN_from_tau_peak(tau0, b, wa0, osc), b
